import pytest
from upgini.metadata import ModelTaskType, RuntimeParameters
from upgini.utils.custom_loss_utils import get_runtime_params_custom_loss, get_additional_params_custom_loss


@pytest.mark.parametrize(
    "input",
    [
        ("regression", ModelTaskType.REGRESSION, True),
        ("poisson", ModelTaskType.REGRESSION, True),
        ("binary", ModelTaskType.BINARY, True),
        ("binary", ModelTaskType.REGRESSION, False),
        ("multiclass", ModelTaskType.MULTICLASS, True),
        ("multiclass_ova", ModelTaskType.BINARY, False),
    ],
)
def test_get_runtime_params_custom_loss(input):
    loss, model_task_type, result = input
    runtime_parameters = RuntimeParameters()
    runtime_parameters.properties["key"] = "value"
    runtime_parameters = get_runtime_params_custom_loss(loss, model_task_type, runtime_parameters, logger=None)
    if result:
        assert runtime_parameters.properties == {
            "key": "value",
            "lightgbm_params_preselection.objective": loss,
            "lightgbm_params_base.objective": loss,
            "lightgbm_params_segment.objective": loss,
        }
    else:
        assert runtime_parameters.properties == {"key": "value"}


@pytest.mark.parametrize(
    "input",
    [
        ("regression", "RMSE", ModelTaskType.REGRESSION, True),
        ("poisson", "Poisson", ModelTaskType.REGRESSION, True),
        ("binary", "Logloss", ModelTaskType.BINARY, True),
        ("binary", "", ModelTaskType.REGRESSION, False),
        ("multiclass", "MultiClass", ModelTaskType.MULTICLASS, True),
        ("multiclass_ova", "MultiClassOneVsAll", ModelTaskType.BINARY, False),
    ],
)
def test_get_additional_params_custom_loss(input):
    loss, loss_cb, model_task_type, result = input
    params = get_additional_params_custom_loss(loss, model_task_type, logger=None)
    if result:
        assert params == {"loss_function": loss_cb}
    else:
        assert params == {}



from functools import reduce
from typing import Any, Dict, List, Optional, Tuple, Union
import numpy as np

import pandas as pd
from sklearn.model_selection import BaseCrossValidator, KFold, TimeSeriesSplit, GroupKFold, GroupShuffleSplit

from upgini.metadata import CVType
from upgini.utils.blocked_time_series import BlockedTimeSeriesSplit


class CVConfig:
    def __init__(
        self,
        cv_type: Union[CVType, str, None],
        date_column: Optional[pd.Series],
        random_state=None,
        shuffle_kfold: Optional[bool] = None,
        test_size: Optional[float] = 0.2,
        n_folds: Optional[int] = 5,
        group_columns: Optional[List[str]] = None,
    ):
        if cv_type is None:
            self.cv_type = CVType.k_fold
        elif isinstance(cv_type, str):
            self.cv_type = CVType(cv_type)
        elif isinstance(cv_type, CVType):
            self.cv_type = cv_type
        else:
            raise Exception(f"Unexpected type of cv_type: {type(cv_type)}")

        self.group_columns = group_columns
        self.shuffle_kfold: Optional[bool] = shuffle_kfold
        self.test_size = test_size
        self.n_folds = n_folds
        if (self.cv_type == CVType.k_fold or self.cv_type == CVType.group_k_fold) and self.shuffle_kfold is None:
            self.shuffle_kfold = date_column is None or is_constant(date_column)
        if self.shuffle_kfold:
            self.random_state = random_state
        else:
            self.random_state = None

    def to_dict(self) -> Dict[str, Any]:
        config = {
            "cv_type": self.cv_type.value,
            "n_folds": self.n_folds,
        }
        if self.shuffle_kfold is not None:
            config["shuffle_kfold"] = self.shuffle_kfold
        if self.test_size is not None:
            config["test_size"] = self.test_size
        return config

    def get_cv_and_groups(self, X: pd.DataFrame) -> Tuple[BaseCrossValidator, Optional[np.ndarray]]:
        if self.cv_type == CVType.time_series:
            return TimeSeriesSplit(n_splits=self.n_folds), None
        elif self.cv_type == CVType.blocked_time_series:
            return BlockedTimeSeriesSplit(n_splits=self.n_folds, test_size=self.test_size), None
        elif self.cv_type == CVType.group_k_fold and self.group_columns:
            groups = get_groups(X, self.group_columns)

            if groups is None or np.unique(groups).size < self.n_folds:
                return KFold(n_splits=self.n_folds, shuffle=self.shuffle_kfold, random_state=self.random_state), None

            if self.shuffle_kfold:
                return (
                    GroupShuffleSplit(n_splits=self.n_folds, test_size=self.test_size, random_state=self.random_state),
                    groups,
                )
            else:
                return GroupKFold(n_splits=self.n_folds), groups
        else:
            return KFold(n_splits=self.n_folds, shuffle=self.shuffle_kfold, random_state=self.random_state), None


def get_groups(X: pd.DataFrame, group_columns: Optional[List[str]]) -> Optional[np.ndarray]:
    existing_group_columns = [c for c in group_columns if c in X.columns]
    return (
        None
        if not group_columns
        else reduce(
            lambda left, right: left + "_" + right, [X[c].astype(str) for c in existing_group_columns]
        ).factorize()[0]
    )


def is_constant(s, dropna=True) -> bool:
    if dropna:
        s = s[pd.notna(s)]
    a = s.to_numpy()
    return True if a.size == 0 else bool((a[0] == a).all())



from upgini.autofe.operand import PandasOperand, VectorizableMixin
import numpy as np
import pandas as pd
from numpy import dot
from numpy.linalg import norm


class Min(PandasOperand):
    name = "min"
    is_binary = True
    has_symmetry_importance = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return np.minimum(left, right)


class Max(PandasOperand):
    name = "max"
    is_binary = True
    has_symmetry_importance = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return np.maximum(left, right)


class Add(PandasOperand, VectorizableMixin):
    name = "+"
    alias = "add"
    is_binary = True
    has_symmetry_importance = True
    is_vectorizable = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return left + right

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        group_column, value_columns = self.validate_calculation(data.columns, **kwargs)
        d1 = data[value_columns]
        d2 = data[group_column]

        return d1.add(d2, axis=0)


class Subtract(PandasOperand, VectorizableMixin):
    name = "-"
    alias = "sub"
    is_binary = True
    has_symmetry_importance = True
    is_vectorizable = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return left - right

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        group_column, value_columns = self.validate_calculation(data.columns, **kwargs)
        d1 = data[value_columns]
        d2 = data[group_column]

        return d1.sub(d2, axis=0)


class Multiply(PandasOperand, VectorizableMixin):
    name = "*"
    alias = "mul"
    is_binary = True
    has_symmetry_importance = True
    is_vectorizable = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return left * right

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        group_column, value_columns = self.validate_calculation(data.columns, **kwargs)
        d1 = data[value_columns]
        d2 = data[group_column]

        return d1.mul(d2, axis=0)


class Divide(PandasOperand, VectorizableMixin):
    name = "/"
    alias = "div"
    is_binary = True
    has_symmetry_importance = True
    is_vectorizable = True
    output_type = "float"

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return left / right.replace(0, np.nan)

    def calculate_group(self, data: pd.DataFrame, **kwargs) -> pd.DataFrame:
        group_column, value_columns = self.validate_calculation(data.columns, **kwargs)
        d1 = data[value_columns]
        d2 = data[group_column]

        return d1.div(d2.replace(0, np.nan), axis=0)


class Combine(PandasOperand):
    name = "Combine"
    is_binary = True
    has_symmetry_importance = True
    output_type = "object"

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        temp = left.astype(str) + "_" + right.astype(str)
        temp[left.isna() | right.isna()] = np.nan
        return pd.Series(temp, index=left.index)


class CombineThenFreq(PandasOperand):
    name = "CombineThenFreq"
    is_binary = True
    has_symmetry_importance = True
    output_type = "float"
    is_distribution_dependent = True
    input_type = "discrete"

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        temp = left.astype(str) + "_" + right.astype(str)
        temp[left.isna() | right.isna()] = np.nan
        value_counts = temp.value_counts(normalize=True)
        self._loc(temp, value_counts)


class Sim(PandasOperand):
    name = "sim"
    is_binary = True
    output_type = "float"
    has_symmetry_importance = True

    def calculate_binary(self, left: pd.Series, right: pd.Series) -> pd.Series:
        return dot(left, right) / (norm(left) * norm(right))



import logging
from logging import Logger
from typing import List, Optional

import pandas as pd
from pandas.api.types import is_object_dtype, is_integer_dtype, is_string_dtype
from upgini.resource_bundle import bundle
from upgini.utils.warning_counter import WarningCounter


class FeaturesValidator:
    def __init__(self, logger: Optional[Logger] = None):
        if logger is not None:
            self.logger = logger
        else:
            self.logger = logging.getLogger()
            self.logger.setLevel("FATAL")

    def validate(self, df: pd.DataFrame, features: List[str], warning_counter: WarningCounter) -> List[str]:
        # one_hot_encoded_features = []
        empty_or_constant_features = []

        for f in features:
            column = df[f]
            if is_object_dtype(column):
                column = column.astype("string")
            value_counts = column.value_counts(dropna=False, normalize=True)
            most_frequent_percent = value_counts.iloc[0]

            if most_frequent_percent >= 0.99:
                empty_or_constant_features.append(f)

            # TODO implement one-hot encoding check
            # if len(value_counts) == 1:
            #     empty_or_constant_features.append(f)
            # elif most_frequent_percent >= 0.99:
            #     empty_or_constant_features.append(f)
            #     if set(value_counts.index.to_list()) == {0, 1}:
            #         one_hot_encoded_features.append(f)
            #     else:
            #         empty_or_constant_features.append(f)
            #     continue

        # if one_hot_encoded_features:
        #     msg = bundle.get("one_hot_encoded_features").format(one_hot_encoded_features)
        #     print(msg)
        #     self.logger.warning(msg)
        #     warning_counter.increment()

        if empty_or_constant_features:
            msg = bundle.get("empty_or_contant_features").format(empty_or_constant_features)
            print(msg)
            self.logger.warning(msg)

        return empty_or_constant_features

    @staticmethod
    def find_high_cardinality(df: pd.DataFrame) -> List[str]:
        # Remove high cardinality columns
        row_count = df.shape[0]
        return [
            i
            for i in df
            if (is_string_dtype(df[i]) or is_integer_dtype(df[i])) and (df[i].nunique() / row_count >= 0.9)
        ]

    @staticmethod
    def find_constant_features(df: pd.DataFrame) -> List[str]:
        return [
            i
            for i in df
            if df[i].nunique() == 1
        ]



import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from upgini.errors import ValidationError
from upgini.metadata import SYSTEM_RECORD_ID, TARGET, ModelTaskType
from upgini.resource_bundle import bundle
from upgini.utils.target_utils import balance_undersample, define_task


def test_invalid_target():
    y = pd.Series(["", "", ""])
    with pytest.raises(ValidationError, match=bundle.get("empty_target")):
        define_task(y)

    y = pd.Series([np.nan, np.inf, -np.inf])
    with pytest.raises(ValidationError, match=bundle.get("empty_target")):
        define_task(y)

    y = pd.Series([1, 1, 1, 1, 1])
    with pytest.raises(ValidationError, match=bundle.get("dataset_constant_target")):
        define_task(y)


def test_define_binary_task_type():
    y = pd.Series([0, 1, 0, 1, 0, 1])
    assert define_task(y, False) == ModelTaskType.BINARY
    assert define_task(y, True) == ModelTaskType.BINARY

    y = pd.Series(["a", "b", "a", "b", "a"])
    assert define_task(y, False) == ModelTaskType.BINARY
    assert define_task(y, True) == ModelTaskType.BINARY


def test_define_multiclass_task_type():
    y = pd.Series(range(1, 51))
    assert define_task(y, False) == ModelTaskType.MULTICLASS
    assert define_task(y, True) == ModelTaskType.MULTICLASS

    y = pd.Series([float(x) for x in range(1, 51)])
    assert define_task(y, False) == ModelTaskType.MULTICLASS
    assert define_task(y, True) == ModelTaskType.MULTICLASS

    y = pd.Series(range(0, 50))
    assert define_task(y, False) == ModelTaskType.MULTICLASS
    assert define_task(y, True) == ModelTaskType.MULTICLASS

    y = pd.Series(["a", "b", "c", "b", "a"])
    assert define_task(y, False) == ModelTaskType.MULTICLASS
    assert define_task(y, True) == ModelTaskType.MULTICLASS

    y = pd.Series(["0", "1", "2", "3", "a"])
    assert define_task(y, False) == ModelTaskType.MULTICLASS
    assert define_task(y, True) == ModelTaskType.MULTICLASS

    y = pd.Series([0.0, 3.0, 5.0, 0.0, 5.0, 0.0, 3.0, 0.0, 5.0, 0.0, 5.0, 0.0, 3.0, 0.0, 3.0, 5.0, 3.0])
    assert define_task(y, False) == ModelTaskType.MULTICLASS


def test_define_regression_task_type():
    y = pd.Series([0.0, 3.0, 5.0, 0.0, 5.0, 0.0, 3.0, 0.0, 5.0, 0.0, 5.0, 0.0, 3.0, 0.0, 3.0, 5.0, 3.0])
    assert define_task(y, True) == ModelTaskType.REGRESSION

    y = pd.Series([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.5])
    assert define_task(y, False) == ModelTaskType.REGRESSION
    assert define_task(y, True) == ModelTaskType.REGRESSION

    y = pd.Series([0, 1, 2, 3, 4, 5, 6, 8])
    assert define_task(y, False) == ModelTaskType.REGRESSION
    assert define_task(y, True) == ModelTaskType.REGRESSION

    y = pd.Series([0.0, 3.0, 5.0, 0.0, 5.0, 0.0, 3.0])
    assert define_task(y, False) == ModelTaskType.REGRESSION
    assert define_task(y, True) == ModelTaskType.REGRESSION


def test_balance_undersampling_binary():
    df = pd.DataFrame({SYSTEM_RECORD_ID: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], TARGET: [0, 1, 0, 0, 0, 0, 0, 0, 0, 0]})
    balanced_df = balance_undersample(
        df, TARGET, ModelTaskType.BINARY, 42, imbalance_threshold=0.1, min_sample_threshold=2
    )
    # Get all minority class and 5x of majority class if minority class count (1)
    # more or equal to min_sample_threshold/2 (1)
    expected_df = pd.DataFrame({
        SYSTEM_RECORD_ID: [1, 2, 3, 7, 9, 10],
        TARGET: [0, 1, 0, 0, 0, 0]
    })
    assert_frame_equal(balanced_df.sort_values(by=SYSTEM_RECORD_ID).reset_index(drop=True), expected_df)

    balanced_df = balance_undersample(
        df, TARGET, ModelTaskType.BINARY, 42, imbalance_threshold=0.1, min_sample_threshold=8
    )
    # Get all minority class and fill up to min_sample_threshold (8) by majority class
    expected_df = pd.DataFrame({
        SYSTEM_RECORD_ID: [1, 2, 3, 4, 6, 7, 9, 10],
        TARGET: [0, 1, 0, 0, 0, 0, 0, 0]
    })
    assert_frame_equal(balanced_df.sort_values(by=SYSTEM_RECORD_ID).reset_index(drop=True), expected_df)

    df = pd.DataFrame({"system_record_id": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], TARGET: [0, 1, 0, 0, 0, 0, 0, 0, 1, 0]})
    balanced_df = balance_undersample(
        df, "target", ModelTaskType.BINARY, 42, imbalance_threshold=0.1, min_sample_threshold=4
    )
    # Get full dataset if majority class count (8) less than x5 of minority class count (2)
    assert_frame_equal(balanced_df, df)


def test_balance_undersaampling_multiclass():
    df = pd.DataFrame({
        SYSTEM_RECORD_ID: [1, 2, 3, 4, 5, 6],
        TARGET: ["a", "b", "c", "c", "b", "c"]
        # a - 1, b - 2, c - 3
    })
    balanced_df = balance_undersample(
        df, TARGET, ModelTaskType.MULTICLASS, 42, imbalance_threshold=0.1, min_sample_threshold=10
    )
    # Get full dataset if majority class count (3) less than x2 of 25% class (b) count (2)
    assert_frame_equal(balanced_df, df)

    df = pd.DataFrame({
        SYSTEM_RECORD_ID: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        TARGET: ["a", "b", "c", "c", "c", "b", "c", "d", "d", "d", "c"]
        # a - 1, b - 2, c - 5, d - 3
    })
    balanced_df = balance_undersample(
        df, TARGET, ModelTaskType.MULTICLASS, 42, imbalance_threshold=0.1, min_sample_threshold=10
    )
    expected_df = pd.DataFrame({
        SYSTEM_RECORD_ID: [1, 2, 3, 4, 5, 6, 8, 9, 10, 11],
        TARGET: ["a", "b", "c", "c", "c", "b", "d", "d", "d", "c"]
    })
    # Get all of 25% quantile class (b) and minor classes (a) and x2 (or all if less) of major classes
    assert_frame_equal(balanced_df.sort_values(by=SYSTEM_RECORD_ID).reset_index(drop=True), expected_df)
